# -*- coding: utf-8 -*-
#
# Copyright © 2012,2013 Pôle de compétences EOLE <eole@ac-dijon.fr>
#
# License CeCILL:
#  * in french: http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
#  * in english http://www.cecill.info/licences/Licence_CeCILL_V2-en.html
##########################################################################

""" Package management with multiple package manager support

This class allows to manage package installation and removing,
system package update.

    Package management interface supports :
       - apt

    To be done :
       - yum
       - pacman
       - urpmi
"""
import os
import sys
import argparse
import pty
import json
import logging
from glob import glob
import pyudev
import urlparse

from pyeole import scriptargs
from pyeole.eoleversion import EOLE_VERSION
from pyeole.eoleversion import EOLE_RELEASE
from pyeole.log import init_logging
from pyeole.process import system_code, system_out
from pyeole.diagnose import test_http
from pyeole.translation.i18n import _

from creole.client import CreoleClient

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

try:
    import apt
    import apt_pkg
    from aptsources.sourceslist import SourcesList, SourceEntry
except ImportError, err:
    log.debug(err.message)

try:
    import yum
except ImportError:
    log.debug(u'python-yum not found')


_TESTRUN = False

_EOLE_APT_CONF_FILE = u'/etc/apt/apt-eole.conf'
"""EOLE configuration file for apt.

"""

_CONFFILE_DEFAULT_ANSWER = u'y'
"""Default answer if progress installer get an question.
"""

_DEBIAN_FRONTEND = u'noninteractive'
"""Make dpkg and debconf completely silent
"""

_SOURCES_CONF = {u'ubuntu': {u'mirror': u'http://eole.ac-dijon.fr/ubuntu',
                             u'sources_file': u'/etc/apt/sources.list',
                             u'level': {u'proposed': [u'stable', u'security', u'updates'],
                                        u'updates': [u'stable', u'security'],
                                        u'security': [u'stable']}},
                 u'eole': {u'mirror': u'http://eole.ac-dijon.fr/eole',
                           u'sources_file': u'/etc/apt/sources.list.d/eole.list',
                           u'level': {u'proposed-updates': [u'stable', u'security', u'updates'],
                                      u'updates': [u'stable', u'security'],
                                      u'security': [u'stable']}},
                 u'cdrom': {u'mirror': u'file:/cdrom/',
                            u'sources_file':  u'/etc/apt/sources.list.d/eole.list',
                            u'level': {u'proposed-updates': None,
                                       u'updates': None,
                                       u'security': None,
                                       u'stable': None}}
                }

_ARCHIVE_DIR = '/var/cache/apt/archives/'

_MOUNT_POINT = '/media/cdrom'

class PackageError(StandardError):
    """Base exception for package management

    """
    pass

class PackageNotFoundError(PackageError):
    """Requested package is not found

    """
    pass


class RepositoryError(PackageError):
    """Unable to reach APT repository

    """
    pass


def parse_apt_cmdline():
    """Parse commande line.
    """
    ubuntu_components = [u'main', u'universe', u'multiverse', u'restricted']

    configure_args = [u'mirror', u'vendor', u'dist', u'level', u'component']
    pkg_args = [u'packages']

    eole_components = [u'main']

    # Global parser
    parser = argparse.ArgumentParser(description=u'Run apt in non-interactive',
                                     parents=[scriptargs.container(),
                                              scriptargs.logging('info')])

    parser.add_argument(u'-j', u'--json', action=u'store_true',
                        help=u'Output in JSON format')

    parser.add_argument(u'-s', u'--simulate', u'--just-print',
                        u'--dry-run', u'--recon', u'--no-act',
                        action=u'store_true',
                        help=u'No action; perform a simulation')

    # per action subparser
    actions = parser.add_subparsers(dest=u'action',
                                    title=u'APT actions',
                                    help=u'actions help')

    # Configuration of sources.list
    configure = actions.add_parser(u'configure',
                                   help=u'configure sources.list')
    vendor = configure.add_subparsers(dest=u'vendor',
                                      title=u'Vendor sources.list configuration',
                                      help=u'distribution vendor')

    ubuntu = vendor.add_parser(u'ubuntu', help=u'Ubuntu sources.list')
    ubuntu.add_argument(u'-m', u'--mirror', metavar=u'URI', type=str,
                        default=u'http://eoleng.ac-dijon.fr/ubuntu',
                        help=u'URI of the mirror')
    ubuntu.add_argument(u'-d', u'--dist', type=str,
                        default=u'precise', help=u'distribution name')
    ubuntu.add_argument(u'-l', u'--level',
                        choices=_SOURCES_CONF[u'ubuntu'][u'level'].keys(),
                        default=u'updates',
                        help=u'update level')
    ubuntu.add_argument(u'-c', u'--component', nargs=u'+',
                        choices=ubuntu_components,
                        default=ubuntu_components,
                        help=u'component to activate')
    ubuntu.set_defaults(sources_file=u'/etc/apt/sources.list')
    ubuntu.set_defaults(func=u'configure')
    ubuntu.set_defaults(func_args=configure_args)

    eole = vendor.add_parser(u'eole', help=u'EOLE sources.list')
    eole.add_argument(u'-m', u'--mirror', metavar=u'URI', type=str,
                      default=u'http://eole.ac-dijon.fr/eole',
                      help=u'URI of the mirror')

    eole.add_argument(u'-d', u'--dist', type=str,
                      choices=[u'stable', u'testing', u'unstable'],
                      default=u'stable', help=u'distribution name')

    eole.add_argument(u'-c', u'--component', nargs=u'+',
                        choices=eole_components,
                        default=eole_components,
                        help=u'component to activate')

    eole.set_defaults(sources_file=u'/etc/apt/sources.list.d/eole.list')
    eole.set_defaults(func=u'configure')
    eole.set_defaults(func_args=configure_args)

    update = actions.add_parser(u'update', help=u'update list of indexes')
    update.set_defaults(func=u'update')

    upgrade = actions.add_parser(u'upgrade', help=u'upgrade packages')
    upgrade.set_defaults(func=u'upgrade')

    dist_upgrade = actions.add_parser(u'dist-upgrade', help=u'upgrade distribution')
    dist_upgrade.set_defaults(func=u'dist_upgrade')

    list_upgrade = actions.add_parser(u'list-upgrade', help=u'list packages to upgrade')
    list_upgrade.set_defaults(func=u'list_upgrade')

    install = actions.add_parser(u'install', help=u'install packages')
    install.add_argument(u'packages', metavar=u'package', type=str, nargs=u'+',
                         help=u'package names')
    install.set_defaults(func=u'install')
    install.set_defaults(func_args=pkg_args)

    remove = actions.add_parser(u'remove', help=u'remove packages')
    remove.add_argument(u'packages', metavar=u'package', type=str, nargs=u'+',
                        help=u'package names')
    remove.set_defaults(func=u'remove')
    remove.set_defaults(func_args=pkg_args)

    opts = parser.parse_args()

    if opts.action == u'configure' and opts.vendor == u'eole':
        opts = _prepare_eole_configure(opts, parser)

    if opts.verbose:
        opts.log_level = u'info'
    if opts.debug:
        opts.log_level = u'debug'
    if opts.json:
        opts.log_level = u'critical'

    return opts


def _prepare_eole_configure(options, parser):
    """Prepare a configure for EOLE

    Unstable and testing do not have an update “level”, use “stable”
    to strip it.

    """
    if options.dist == 'stable':
        options.dist = 'eole-{0}'.format(EOLE_RELEASE)
        options.level = 'updates'
    elif options.dist == 'testing':
        options.dist = 'eole-{0}-testing'.format(EOLE_VERSION)
        options.level = 'stable'
    elif options.dist == 'unstable':
        options.dist = 'eole-{0}-unstable'.format(EOLE_VERSION)
        options.level = 'stable'

    return options


def _build_apt_cmd_line(**kwargs):
    """Build APT commande line from arguments

    This is usefull to pass apt-eole calls in containers.

    """
    action = kwargs[u'action']
    cmd = [u'apt-eole']
    if kwargs.get(u'container', None):
        cmd.extend([u'--container', kwargs.get(u'container')])
    if kwargs.get(u'json', False):
        cmd.append(u'--json')
    cmd.extend([u'--log-level', kwargs.get(u'log_level', u'info')])
    cmd.append(action)
    if action == u'configure':
        cmd.append(kwargs[u'vendor'])
        cmd.extend([u'--mirror', kwargs[u'mirror']])
        cmd.extend([u'--level', kwargs[u'level']])
        if kwargs.get(u'component', None):
            cmd.extend([u'--component'] + kwargs[u'component'])
        # cmd.extend([u'--source-file', kwargs[u'sources_files']])
    elif action in [u'install', u'remove']:
        cmd.extend(kwargs[u'packages'])

    return cmd


def apt_cli():
    """Get command line arguments and run the action

    """
    opts = vars(parse_apt_cmdline())
    cli_log = init_logging(name=u'apt-eole', level=opts[u'log_level'],
                           as_root=True)
    try:
        pkgmgr = EolePkg(u'apt')
        if opts.get(u'simulate', False):
            pkgmgr.set_option(u'APT::Get::Simulate', u'true')
        if opts.get(u'json', False):
            pkgmgr.set_json_output(root=True)
        func_args = {}
        for arg_name in opts.get(u'func_args', []):
            if opts.get(arg_name, None) is None:
                continue
            func_args[arg_name] = opts[arg_name]

        getattr(pkgmgr, opts[u'func'])(container=opts[u'container'], **func_args)
    except Exception, err:
        if opts[u'debug']:
            cli_log.error(err.message, exc_info=True)
        else:
            cli_log.error(err.message)
        sys.exit(1)
    sys.exit(0)


class NonInteractiveInstallProgress(apt.progress.base.InstallProgress):
    """Make installation/remove process silent.
    """

    def __init__(self, silent=False):
        """Initialize the NonInteractiveInstallProgress
        """
        self.silent = silent
        apt.progress.base.InstallProgress.__init__(self)

    def conffile(self, current, new):
        os.write(self.master_fd, "%s\n" % _CONFFILE_DEFAULT_ANSWER)

    def fork(self):
        (pid, self.master_fd) = pty.fork()
        if pid == 0:
            os.environ["DEBIAN_FRONTEND"] = _DEBIAN_FRONTEND
            # rt ticket #49720
            if not "PATH" in os.environ or not "sbin" in os.environ["PATH"]:
                os.environ["PATH"] = "/usr/sbin:/usr/bin:/sbin:/bin"
        return pid

    def update_interface(self):
        try:
            if self.silent:
                # Avoid blocking file descriptor
                os.read(self.master_fd, 512),
            else:
                print os.read(self.master_fd, 512),
        except:
            pass
        apt.progress.base.InstallProgress.update_interface(self)


class EolePkgApt(object):
    """apt wrapper class.
    This class mangage the interaction with "apt" based systems, like debian or ubutnu.
    This is a 'private' class it is made to be used by :class:`EolePkg` with the 'apt'
    parameter

    - prototype: ``EolePkgApt(root)`` with ``root`` the root directory for
    package installation

    When :method:`__init__` is called you initialize a new interface for package
    management.

    """
    def __init__(self, root=None, test=False, type_='http'):
        """Initialize the apt wrapper
        Load end prepare all the necessary to manage packages with `apt`

        :param root: The system directory to be the root for package installation
        :type root: directory `string`
        :raise ImportError: if python-apt is not available.
        """
        self.cache = None
        self._rootdir = root
        self._read_config_file()
        # Progress
        self._acquire_progress = apt.progress.TextFetchProgress()
        self._install_progress = NonInteractiveInstallProgress()
        # silent version
        self._silent_install_progress = NonInteractiveInstallProgress(silent=True)

        self.client = CreoleClient()
        # Do not run in container mode within a container
        self.container_mode = not self.client.is_in_lxc() \
                              and self.client.get_creole(u'mode_conteneur_actif', u'non') == u'oui'

        self._return_json = False

        self.set_test_run(_TESTRUN)
        self.set_test_run(test)


    def set_test_run(self, value):
        """Set a value to :attr:`self._testrun`
        Used by the unit test for dry run, don't use this
        """
        self._testrun = value


    def _load_apt_cache(self, action=None):
        """Load APT cache

        :param action: which action was called
        :type action: `str`

        """
        if self.cache is None:
            try:
                self.cache = apt.Cache(rootdir=self._rootdir)
            except SystemError, err:
                if action != 'configure':
                    raise err
            else:
                #Test dpkg
                if self.cache.dpkg_journal_dirty:
                    raise SystemError(u'dpkg journal is dirty, please run '
                                      '"dpkg --configure -a" and retry it')
                self.cache.open()


    def _run(self, action, func, container=None, silent=False, **kwargs):
        """Run apt action.

        :param silent_apt: avoid apt making outputs on standard output
        :type silent_apt: `bool`

        """
        if self.container_mode and container not in [None, u'all', u'root', u'current']:
            return {}

        if container != u'current':
            log.info(_(u'Running {0} for root').format(action))

        self._load_apt_cache(action)

        return {u'root': getattr(self, func)(silent=silent, **kwargs)}


    def _run_containers(self, action, container, **kwargs):
        """Run apt-eole in containers
        """
        containers = []
        outputs = {}
        if self.container_mode and container != u'current':
            if container == u'all':
                containers = self.client.get_groups()
            elif container not in [None, u'root']:
                containers = [container]

            # Strip unwanted containers
            containers = [ name for name in containers if name not in [u'root', u'all'] ]

            cmd = _build_apt_cmd_line(action=action,
                                      container=u'current',
                                      json=self._return_json,
                                      **kwargs)
            log.debug(u'Run: {0}'.format(' '.join(cmd)))
            env = {u'DEBIAN_FRONTEND':_DEBIAN_FRONTEND}
            for container in containers:
                log.info(_(u'Running {0} for container {1}').format(action,
                                                                 container))
                if self._return_json:
                    code, stdout, stderr = system_out(cmd, container=container,
                                                      env=env)
                else:
                    code = system_code(cmd, container=container, env=env)
                    stdout = None
                    stderr = None

                if code == 0:
                    if self._return_json:
                        # Every container see itself as root
                        outputs[container] = json.loads(stdout)[u'root']
                    else:
                        outputs[container] = stdout
                else:
                    cmd_string = u' '.join(cmd)
                    msg = _(u"Can not execute '{0}' for container '{1}'")
                    if stderr:
                        msg += u": {2}"
                    raise SystemError(msg.format(cmd_string, container, stderr))

        return outputs


    def _read_config_file(self):
        """Read the EOLE configuration file.

        """
        if os.access(_EOLE_APT_CONF_FILE, os.R_OK):
            try:
                apt_pkg.read_config_file(apt_pkg.config, _EOLE_APT_CONF_FILE)
            except SystemError, err:
                log.error(_(u'Unable to load EOLE apt configuration: {0}').format(err))
                log.debug(err.message, exc_info=True)


    def set_option(self, option, value):
        """Set APT option

        """
        apt_pkg.config.set(option, value)


    def commit(self, silent=False):
        """Commit changes and update database view

        The :meth:`cache.commit` method does not update the packages
        informations, we need to reopen it.

        :param silent: do not display progress informations
        :type silent: `bool`

        """
        if self._testrun:
            self.cache.clear()
        else:
            if silent:
                self.cache.commit(install_progress=self._silent_install_progress)
            else:
                self.cache.commit(self._acquire_progress, self._install_progress)
            self.cache.open()


    def get_package(self, pkgname):
        """Search a package in package database.

        :param pkgname: The package name
        :type pkgname: `str`
        :return: a package from the package database.
        :rtype: `apt.Package`

        """
        try:
            self._load_apt_cache()
            return self.cache[pkgname]
        except KeyError:
            raise PackageNotFoundError(_(u'Package {0} not found').format(pkgname))

    def get_rev_depends(self, pkgname):
        """ get reverse depends list

        :param pkgname: The package name
        :type pkgname: `str`
        :return: a package name list.
        :rtype: `list`
        """
        dep_list = []
        self._load_apt_cache()
        pkg = self.cache[pkgname]
        for el in self.cache.keys():
            pkg = self.cache[el]
            deps = pkg.candidateDependencies
            for dep in deps:
                for el in dep:
                    if pkgname  == el.name:
                        dep_list.append(pkg.name)
        return dep_list

    @staticmethod
    def _test_mirror(mirror, dist):
        """Test if mirror is reachable
        """
        host = urlparse.urlparse(mirror).netloc
        proxy_opt = "Acquire::http::Proxy"
        env_proxy = os.environ.get('http_proxy')
        default_proxy = apt_pkg.config.get(proxy_opt)
        host_proxy = apt_pkg.config.get("{0}::{1}".format(proxy_opt, host))
        if not env_proxy:
            if host_proxy:
                os.environ.update({'http_proxy': host_proxy})
            elif default_proxy:
                os.environ.update({'http_proxy': default_proxy})

        release_url = u'{0}/dists/{1}/Release'.format(mirror, dist)
        if not test_http(release_url, use_proxy=True):
            msg = _(u'Unable to get repository Release: {0}')
            raise RepositoryError(msg.format(release_url))
        if not env_proxy:
            # reset environment
            os.environ.pop('http_proxy', None)


    def configure(self, mirror, vendor, dist, level, component=None, silent=False):
        """Configure APT sources.list
        """
        vendor_conf =  _SOURCES_CONF[vendor]
        sources_file = vendor_conf[u'sources_file']
        sources = SourcesList()
        # Empty sources.list
        sources.list = []
        if component is not None:
            if not mirror.startswith(u'http://') and not mirror.startswith(u'file:/'):
                mirror = u'http://{0}'.format(mirror)

            if vendor_conf[u'level'].get(level, None) is not None:
                for suffix in vendor_conf[u'level'][level]:
                    if suffix == u'stable':
                        dist_name = dist
                    else:
                        dist_name = u'{0}-{1}'.format(dist, suffix)
                    if vendor != 'cdrom':
                        self._test_mirror(mirror, dist)
                    msg = "Add source to {0}: deb {1} {2} {3}"
                    log.debug(msg.format(sources_file, mirror, dist_name, ' '.join(component)))
                    sources.add(u'deb', mirror, dist_name, component,
                                pos=len(sources.list), file=sources_file)

            if level == u'stable':
                level_dist = dist
            else:
                level_dist = u'{0}-{1}'.format(dist, level)
            if vendor != 'cdrom':
                self._test_mirror(mirror, level_dist)
            msg = "Add source to {0}: deb {1} {2} {3}"
            log.debug(msg.format(sources_file, mirror, level_dist, ' '.join(component)))
            sources.add(u'deb', mirror, level_dist, component, pos=len(sources.list),
                        file=sources_file)
        # Save sources.list even if empty
        sources.save()
        # Reload cache if required
        self._load_apt_cache()
        self.cache._list.read_main_list()


    def _mount_cdrom(self):
        """Mount installer CDROM

        Use first CDROM with a “dists/” top level directory.
        """
        if not os.path.ismount(_MOUNT_POINT):
            context = pyudev.Context()
            try:
                cdroms = context.list_devices(ID_CDROM=1)
            except:
                raise SystemError('unable to find CDROM')
            for cdrom in cdroms:
                if system_code(['/bin/mount', cdrom['DEVNAME'], _MOUNT_POINT, '-o', 'ro']) != 0:
                    continue
                elif not os.path.isdir('{0}/dists'.format(_MOUNT_POINT)):
                    self._umount_cdrom()

            if not os.path.ismount(_MOUNT_POINT):
                raise SystemError('No installer CDROM found')
        else:
            log.debug("{0} already mount".format(_MOUNT_POINT))

    def _umount_cdrom(self):
        if os.path.ismount(_MOUNT_POINT):
            if system_code(['/bin/umount', _MOUNT_POINT]) != 0:
                raise SystemError('unable to umount {0}'.format(_MOUNT_POINT))


    def _prepare_cache(self):
        for cache in self.cache._list.list:
            if cache.uri.startswith('file:'+_MOUNT_POINT):
                self._mount_cdrom()
                return


    def get_upgradable_list(self, silent=False):
        """Search all packages who needs to be updated.

        :return: upgradable packages
        :rtype: `list` of `apt.Package`

        """
        packages = []
        with self.cache.actiongroup():
            self.update(silent=True)
            self.cache.upgrade(dist_upgrade=True)

            for package in self.cache.get_changes():
                packages.append([package.name, package.isInstalled])
            # Do not perform action
            self.cache.clear()

        return packages


    def _apply_to_package(self, packages, mark, action, silent=False):
        """Perform an action on a package or package list.

        :param packages: package name
        :type  packages: `str` or `list` of `str`
        :param mark: name of the mark function to apply on the package
        :type mark: `str`
        :param action: name of the action to perform on the package
        :type action: `str`
        :param silent: do not display progress informations
        :type silent: `bool`

        """
        check = {u'install': True, u'remove': False}
        options = {u'install': {},
                   u'remove': {'purge': True}}

        self._prepare_cache()
        if not isinstance(packages, list):
            packages = [packages]

        with self.cache.actiongroup():
            for name in packages:
                pkg = self.get_package(name)
                # Mark the package for the action
                if pkg.isInstalled != check[action]:
                    try:
                        getattr(pkg, mark)(**options[action])
                    except SystemError, err:
                        msg = _(u'Unable to mark package {0} for {1}: {2}')
                        raise PackageError(msg.format(name, action,
                                                      err.message))
            changes = self.cache.get_changes()
            if len(changes) == 0:
                log.info(_(u'No package to {0}.').format(action))
            else:
                to_install = []
                to_remove = []
                to_upgrade = []
                to_downgrade = []
                for pkg in sorted(changes):
                    if pkg.marked_install:
                        to_install.append(pkg.name)
                    elif pkg.marked_delete:
                        to_remove.append(pkg.name)
                    elif pkg.marked_upgrade:
                        to_upgrade.append(pkg.name)
                    elif pkg.marked_downgrade:
                        to_downgrade.append(pkg.name)

                if to_install:
                    msg = _(u'Install: {0}.')
                    log.info(msg.format(u', '.join(to_install)))
                if to_remove:
                    msg = _(u'Remove: {0}.')
                    log.info(msg.format(u', '.join(to_remove)))
                if to_upgrade:
                    msg = _(u'Upgrade: {0}.')
                    log.info(msg.format(u', '.join(to_upgrade)))
                if to_downgrade:
                    msg = _(u'Downgrade: {0}.')
                    log.info(msg.format(u', '.join(to_downgrade)))

            self.commit(silent=silent)
        self._umount_cdrom()


    def install(self, packages, silent=False):
        """Install a package or a list of packages

        :param packages: package name
        :type  packages: `str` or `list` of `str`
        :param silent: do not display progress informations
        :type silent: `bool`

        """
        self._apply_to_package(packages, u'mark_install', u'install',
                               silent=silent)


    def remove(self, packages, silent=False):
        """Remove a package

        :param packages: package name
        :type  packages: `str` or `list` of `str`
        :param silent: do not display progress informations
        :type silent: `bool`

        """
        self._apply_to_package(packages, u'mark_delete', u'remove',
                               silent=silent)


    def update(self, silent=False):
        """ update the package lists (same as apt-get update)

        :param silent: do not display progress informations
        :type silent: `bool`

        """
        self._prepare_cache()
        if silent:
            self.cache.update()
        else:
            self.cache.update(self._acquire_progress)
        self.cache.open()
        self._umount_cdrom()


    def upgrade(self, dist_upgrade=False, silent=False):
        """ upgrade all packages with new available versions (same as apt-get upgrade)

        :param silent: do not display progress informations
        :type silent: `bool`

        """
        self._prepare_cache()
        self.cache.upgrade(dist_upgrade=dist_upgrade)
        self.commit(silent=silent)
        self._umount_cdrom()


    def dist_upgrade(self, silent=False):
        """make dist-upgrade (same as apt-get dist-upgrade)

        :param silent: do not display progress informations
        :type silent: `bool`

        """
        self.upgrade(dist_upgrade=True, silent=silent)


    def fetch_archives(self, packages, silent=False):
        self._prepare_cache()
        with self.cache.actiongroup():
            for package in packages:
                pkg = self.get_package(package)
                if pkg.is_upgradable:
                    pkg.mark_install()
        self.cache.fetch_archives()
        self._umount_cdrom()

    def get_depends(self, pkgname):
        """Search dependencies for a package

        :param pkgname: package name
        :type  pkgname: `str`
        :return: dependencies for a package
        :rtype: `list` of `apt.Package`
        """
        depends = []

        pkg = self.get_package(pkgname)
        candidates = pkg.candidate.dependencies

        for dep in candidates:
            for elm in dep.or_dependencies:
                depends.append(elm.name)

        return depends


    def download(self, pkgname, dest=u'/tmp', silent=False):
        """Download a package binary

        The default destination is ``/tmp``.

        :param pkgname: package name
        :type  pkgname: `str`
        :param dest: destination directory
        :type dest: `str`
        :param silent: do not display progress informations
        :type silent: `bool`

        """
        self._prepare_cache()
        pkg = self.cache[pkgname]
        log.info(_(u'Downloading package {0}').format(pkgname))
        if silent:
            pkg.candidate.fetch_binary(dest)
        else:
            pkg.candidate.fetch_binary(dest, progress=self._acquire_progress)
        self._umount_cdrom()

    def clean(self):
        """Clean old deb file in cache dir

        see cmdline/apt-get.cc DoClean
        and apt-pkg/acquire.cc pkgAcquire::Clean
        """
        exclude_file = ['lock', 'partial']
        for dirname in [_ARCHIVE_DIR, os.path.join(_ARCHIVE_DIR, 'partial')]:
            if not os.path.isdir(dirname):
                return
            for filename in glob(os.path.join(dirname, '*')):
                if os.path.basename(filename) in exclude_file:
                    continue
                os.unlink(filename)


class EolePkgYum(object):
    """yum wrapper class.
    This class mangage the interaction with "yum" based systems, like RHEL, CentOS or Fedora
    This is a 'private' class it is made to be used by :class:`EolePkg` with the 'yum'
    parameter

    - prototype: ``EolePkgYum(root)`` with ``root`` the root directory for
    package installation

    When :method:`__init__` is called you initialize a new interface for package
    management.

    """
    def __init__(self, root=None):
        raise NotImplementedError(_(u'  Yum Support is not Implemented yet'))


class EolePkg(object):
    """Main Eole package manipulation class
    Use this class to manipulate packages by passing the type to ``EolePkg()``

    This class manage automaticaly  the packages with the good backend.
    """
    def __init__(self, mode, root=None, test=False, type_='http'):
        """ Initialize the package manager interface
        :param mode: The backend type (apt, yum, urpmi)
        :type mode: string
        :param root: The system directory to be the root for package installation
        :type root: directory `string`
        :param test: used for dry run mode
        :type test: boolean
        """
        if mode == u'apt':
            self.pkgmgr = EolePkgApt(root=root, test=test, type_=type_)
        else:
            raise ImportError(_(u'No package manager module found'))

        self._return_json = False

        self.set_test_run(_TESTRUN)
        self.set_test_run(test)


    def set_test_run(self, value):
        """Set a value to :attr:`self._testrun`
        Used by the unit test for dry run, don't use this
        """
        self._testrun = value
        self.pkgmgr.set_test_run(value)


    def set_option(self, option, value):
        """Set package manager option

        """
        self.pkgmgr.set_option(option, value)


    def set_json_output(self, root=None, container=None):
        """Use to get output of containers as json

        """
        # Return json only if asked on command line
        if root is not None:
            self._return_json = root
        if container is not None:
            self.pkgmgr._return_json = container


    def _run(self, action, func, container=None, **kwargs):
        """Run package manager actions for all containers

        Print a JSON dump of return value if asked.

        :param action: action to perform
        :type action: `str`
        :param func: package manager method to run
        :type func: `str`
        :param container: container name
        :type container: `str`
        :return: return values of commands by containers
        :rtype: `dict`

        """
        ret = {}

        # Run for the current container
        ret.update(self.pkgmgr._run(action=action,
                                    func=func,
                                    container=container,
                                    **kwargs))

        # Try for containers if applicable
        ret.update(self.pkgmgr._run_containers(action=action,
                                               container=container,
                                               **kwargs))

        if self._return_json:
            print json.dumps(ret)
        else:
            return ret


    def get_package(self, pkgname):
        """ get package search the package in package database.
        :param pkgname: package name
        :type: pkgname: `str`
        :return: a package from the package database.
        :rtype: `apt.Package` or `yum.Package`

        """
        return self.pkgmgr.get_package(pkgname)

    def get_rev_depends(self, pkgname):
        """ get package reverse depends list
        :param pkgname:  package name
        :type: pkgname: `str`
        :return: a package name list.
        :rtype: `list`
        """
        return self.pkgmgr.get_rev_depends(pkgname)


    def configure(self, mirror, vendor, dist, level, component=None, container=None, silent=False):
        """Configure package manager.

        """
        action = 'configure'
        func = 'configure'

        if container is None:
            container = u'all'

        return self._run(action=action,
                         func=func,
                         container=container,
                         mirror=mirror,
                         vendor=vendor,
                         dist=dist,
                         level=level,
                         component=component)

    def install(self, packages, container=None, silent=False):
        """Install a package

        :param packages: package names
        :type  packages: `str` or `list` of `str`
        :param silent: do not display progress informations
        :type silent: `bool`

        """
        action = u'install'
        func = u'install'

        if not isinstance(packages, list):
            packages = [packages]

        return self._run(action=action,
                         func=func,
                         packages=packages,
                         container=container,
                         silent=silent)


    def remove(self, packages, container=None, silent=False):
        """Remove a package

        :param packages: package names
        :type  packages: `str` or `list` of `str`
        :param silent: do not display progress informations
        :type silent: `bool`

        """
        action = u'remove'
        func = u'remove'

        if not isinstance(packages, list):
            packages = [packages]

        return self._run(action=action,
                         func=func,
                         packages=packages,
                         container=container,
                         silent=silent)


    def update(self, container=None, silent=False):
        """Update the package lists

        :param silent: do not display progress informations
        :type silent: `bool`

        """
        action = u'update'
        func = u'update'

        if container is None:
            container = u'all'

        return self._run(action=action,
                         func=func,
                         container=container,
                         silent=silent)


    def upgrade(self, container=None, dist=False, silent=False):
        """Upgrade all packages with new available versions

        :param silent: do not display progress informations
        :type silent: `bool`

        """
        func = u'upgrade'
        action = u'upgrade'
        if dist:
            action = u'dist-upgrade'

        if container is None:
            container = u'all'

        return self._run(action=action,
                         func=func,
                         container=container,
                         dist_upgrade=dist,
                         silent=silent)


    def dist_upgrade(self, container=None, silent=False):
        """Upgrade the distribution

        :param silent: do not display progress informations
        :type silent: `bool`

        """
        if container is None:
            container = u'all'

        return self.upgrade(container=container,
                            dist=True,
                            silent=silent)

    def fetch_archives(self, packages, container=None, silent=False):
        if container is None:
            container = u'all'

        func = u'fetch_archives'
        action = 'fetch_archives'

        if container is None:
            container = u'all'

        return self._run(action=action,
                         func=func,
                         packages=packages,
                         container=container,
                         silent=silent)


    def get_depends(self, pkgname=None):
        """Search dependencies for a package

        :param pkgname: package name
        :type  pkgname: `str`
        :return: dependencies for a package
        :rtype: `list` of `apt.Package` or `yum.Package`
        """
        return self.pkgmgr.get_depends(pkgname)


    def download(self, pkgname, dest=u'/tmp', silent=False):
        """Download a package binary

        The default destination is ``/tmp``.

        :param pkgname: package name
        :type  pkgname: `str`
        :param dest: destination directory
        :type dest: `str`
        :param silent: do not display progress informations
        :type silent: `bool`

        """
        return self.pkgmgr.download(pkgname, dest, silent=silent)


    def get_upgradable_list(self, container=None, silent=False):
        """ get_upgradable_list finds all packages who needs to be updated.
        """
        # Make sure to call the good apt-eole action in containers
        action = u'list-upgrade'
        func = u'get_upgradable_list'
        self.set_json_output(container=True)

        if container is None:
            container = u'all'

        ret = self._run(action=action,
                        func=func,
                        container=container,
                        silent=True)

        self.set_json_output(container=False)
        return ret


    def list_upgrade(self, container=None, upgrades=None, silent=False):
        """List upgradable packages

        :param container: container name
        :type container: `str`
        :param upgrades: packages to upgrade/install, per container
        :type upgrades: `dict`
        :param silent: make APT calls silent
        :type silent: `bool`

        """
        msg = []
        if upgrades is None:
            upgrades = self.get_upgradable_list(container=container,
                                                silent=silent)

        if not isinstance(upgrades, dict):
            return

        combined = {u'new': {}, u'upgrade': {}}
        for container, packages in upgrades.items():
            if not packages:
                continue
            for name, isInstalled in packages:
                if isInstalled:
                    if name not in combined[u'upgrade']:
                        combined[u'upgrade'][name] = [container]
                    else:
                        combined[u'upgrade'][name].append(container)
                else:
                    if name not in combined[u'new']:
                        combined[u'new'][name] = [container]
                    else:
                        combined[u'new'][name].append(container)

        if not combined[u'new'] and not combined[u'upgrade']:
            msg.append(u'No upgradable packages')
        else:
            if combined[u'new']:
                msg.append(u'New packages:')
                for pkg, containers in sorted(combined[u'new'].items()):
                    msg.append(u'    {0} ({1})'.format(pkg, u', '.join(containers)))
            if combined['upgrade']:
                #print u'Upgradable packages:'
                msg.append(_(u'Upgradable packages:'))
                for pkg, containers in sorted(combined[u'upgrade'].items()):
                    msg.append(u'    {0} ({1})'.format(pkg, u', '.join(containers)))
        return msg

    def clean(self):
        """Clean packge in cache"""
        self.pkgmgr.clean()
